# -*- coding: utf-8 -*-
import io
import os

from pkg_resources import DistributionNotFound, get_distribution

# flake8: noqa

from .defaults import DEFAULT_FRKL_JINJA_ENVIRONMENT
from .frkl import (
    Frkl,
    load_object_from_url_or_path,
    load_string_from_url_or_path,
    load_templated_string_from_url_chain,
)
from .frklist import Frklist, FrklistContext
from .helpers import content_from_url, dict_from_url, download_cached_file, get_full_url
from .processors import (
    EnsurePythonObjectProcessor,
    EnsureUrlProcessor,
    FrklProcessor,
    LoadMoreConfigsProcessor,
    UrlAbbrevProcessor,
)
from .utils import VarsType

__author__ = """Markus Binsteiner"""
__email__ = "makkus@posteo.de"

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    try:
        version_file = os.path.join(os.path.dirname(__file__), "version.txt")

        if os.path.exists(version_file):
            with io.open(version_file, encoding="utf-8") as vf:
                __version__ = vf.read()
        else:
            __version__ = "unknown"

    except (Exception):
        pass

    if __version__ is None:
        __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound



# -*- coding: utf-8 -*-

import collections
import logging
import pprint
import sys

import click
from six import string_types

from . import __version__ as VERSION
from .frkl_factory import factory

log = logging.getLogger("frkl")


class Config(object):
    """frkl configuration, holds things like aliases and such."""

    def __init__(self, configuration_dict=None):

        if not configuration_dict:
            configuration_dict = {}

        if isinstance(configuration_dict, dict):
            self.config = configuration_dict
        elif isinstance(configuration_dict, string_types):
            self.config = "XXX"
        else:
            raise Exception(
                "frkl configuration needs to be created using a dict object"
            )


@click.group(invoke_without_command=True)
@click.option("--version", help="the version of frkl you are using", is_flag=True)
@click.pass_context
def cli(ctx, version):
    """Console script for frkl"""

    if version:
        click.echo(VERSION)
        sys.exit(0)

    ctx.obj = {}


@cli.command("print-config")
@click.option(
    "--init",
    "-i",
    multiple=True,
    help="config to bootstrap the frkl object itself, if not provided, config strings need to contain at least one folder with init information, refer to documentation for more info",
)
@click.argument("config", required=False, nargs=-1)
@click.pass_context
def print_config(ctx, init, config):
    if not init:
        frkl_obj = init(config)
    else:
        frkl_obj = factory(init, config)

    result = frkl_obj.process()

    if isinstance(result, collections.Iterable):

        print("")
        print(
            "\n# ----------------------------------------\n".join(
                (pprint.pformat(x) for x in result)
            )
        )
        print("")

    else:
        print("")
        print(result)
        print("")


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
test_frkl
----------------------------------

Tests for `frkl` module.
"""

import pprint

import pytest
from ruamel.yaml import YAML
from frutils.frutils import *  # noqa: F403

from frkl.frkl import *  # noqa: F403
from frkl.frkl_factory import *  # noqa: F403
from frkl.callbacks import *  # noqa: F403
from frkl.processors import *  # noqa: F403

# from click.testing import CliRunner
yaml = YAML()

TEST_DICTS = [
    ({}, {}, {}),
    ({"a": 1}, {"a": 1}, {"a": 1}),
    ({"a": 1}, {"b": 1}, {"a": 1, "b": 1}),
    ({"a": 1}, {"a": 2}, {"a": 2}),
    ({"a": 1, "aa": 11}, {"b": 2, "bb": 22}, {"a": 1, "aa": 11, "b": 2, "bb": 22}),
    ({"a": 1, "aa": 11}, {"b": 2, "aa": 22}, {"a": 1, "b": 2, "aa": 22}),
]

TEST_CONVERT_TO_PYTHON_OBJECT_DICT = [
    {"config": {"a": 1, "b": 2}},
    {"config": {"c": 3, "d": 4}},
]

TEST_CUSTOM_ABBREVS = {"test_abbr1": "https://example.url/folder1/folder2/"}

TEST_REGEXES = {"^start": "replacement", "frkl_expl": "makkus/freckles/examples"}

TEST_REGEX_URLS = [
    ("start_resturl", "replacement_resturl"),
    ("xstart_resturl", "xstart_resturl"),
    ("begin/frkl_expl/end", "begin/makkus/freckles/examples/end"),
    ("start/frkl_expl/end", "replacement/makkus/freckles/examples/end"),
]

TESTFILE_1_CONTENT = """- config:
    a: 1
    b: 2

- config:
    c: 3
    d: 4
"""

TEST_FRKLIZE_DICT_1 = [
    {"vars": {"aa": 11, "bb": 22}, "task": {"task_name": "task1"}},
    {"vars": {"aa": 11, "bb": 22}, "task": {"task_name": "task2"}},
]

TEST_FRKLIZE_1_RESULT = [
    {"task": {"task_name": "task1"}, "vars": {"a": 1, "b": 2, "aa": 11, "bb": 22}},
    {"task": {"task_name": "task2"}, "vars": {"a": 1, "b": 2, "cc": 33, "dd": 44}},
]

TEST_FRKLIZE_1_RESULT_DOUBLE = TEST_FRKLIZE_1_RESULT + TEST_FRKLIZE_1_RESULT

TEST_JINJA_DICT = {"config": {"a": 1, "b": 2, "c": 3, "d": 4}}
TEST_JINJA_URLS = [
    (
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_jinja.yaml"
        ),
        TEST_JINJA_DICT,
        TESTFILE_1_CONTENT,
    )
]

TEST_ENSURE_FAIL_URLS = [
    ("/tmp_does_not_exist/234234234"),
    ("https://raw222.githubusercontent.com/xxxxxxx8888/asdf.yml"),
]

TEST_PROCESSOR_CHAIN_1 = [
    RegexProcessor(**{"regexes": TEST_REGEXES}),
    UrlAbbrevProcessor(**{"abbrevs": TEST_CUSTOM_ABBREVS}),
]
TEST_CHAIN_1_URLS = [
    (
        ["gh:makkus/freckles/examples/quickstart.yml"],
        [
            "https://raw.githubusercontent.com/makkus/freckles/master/examples/quickstart.yml"
        ],
    ),
    (
        ["bb:makkus/freckles/examples/quickstart.yml"],
        ["https://bitbucket.org/makkus/freckles/src/master/examples/quickstart.yml"],
    ),
    (
        ["gh:frkl_expl/quickstart.yml"],
        [
            "https://raw.githubusercontent.com/makkus/freckles/master/examples/quickstart.yml"
        ],
    ),
]

REGEX_CHAIN = [RegexProcessor(**{"regexes": TEST_REGEXES})]
JINJA_CHAIN = [
    EnsureUrlProcessor(),
    Jinja2TemplateProcessor(**{"template_values": TEST_JINJA_DICT}),
]
ABBREV_CHAIN = [UrlAbbrevProcessor(**{"abbrevs": TEST_CUSTOM_ABBREVS})]
ENSURE_URL_CHAIN = [EnsureUrlProcessor()]
ENSURE_PYTHON_CHAIN = [EnsureUrlProcessor(), EnsurePythonObjectProcessor()]

FRKL_INIT_PARAMS = {
    STEM_KEY_NAME: "childs",
    DEFAULT_LEAF_KEY_NAME: "task",
    DEFAULT_LEAF_DEFAULT_KEY_NAME: "task_name",
    DEFAULT_LEAF_KEY_MAP_NAME: "vars",
}
FRKLIZE_CHAIN = [
    EnsureUrlProcessor(),
    EnsurePythonObjectProcessor(),
    LoadMoreConfigsProcessor(),
    FrklProcessor(**FRKL_INIT_PARAMS),
]
ABBREV_FRKLIZE_CHAIN = [
    UrlAbbrevProcessor(),
    EnsureUrlProcessor(),
    EnsurePythonObjectProcessor(),
    LoadMoreConfigsProcessor(),
    FrklProcessor(**FRKL_INIT_PARAMS),
]

PROCESSOR_TESTS = [
    (REGEX_CHAIN, "start_resturl", "unprocessed", ["replacement_resturl"]),
    (REGEX_CHAIN, "xstart_resturl", "unprocessed", ["xstart_resturl"]),
    (
        REGEX_CHAIN,
        "begin/frkl_expl/end",
        "unprocessed",
        ["begin/makkus/freckles/examples/end"],
    ),
    (
        REGEX_CHAIN,
        "start/frkl_expl/end",
        "unprocessed",
        ["replacement/makkus/freckles/examples/end"],
    ),
    (
        ENSURE_URL_CHAIN,
        os.path.join(os.path.dirname(os.path.realpath(__file__)), "testfile.yaml"),
        "unprocessed",
        [TESTFILE_1_CONTENT],
    ),
    (
        ENSURE_URL_CHAIN,
        "https://raw.githubusercontent.com/makkus/frkl/master/tests/testfile.yaml",
        "unprocessed",
        [TESTFILE_1_CONTENT],
    ),
    (
        ENSURE_PYTHON_CHAIN,
        os.path.join(os.path.dirname(os.path.realpath(__file__)), "testfile.yaml"),
        "unprocessed",
        [TEST_CONVERT_TO_PYTHON_OBJECT_DICT],
    ),
    (
        ENSURE_PYTHON_CHAIN,
        "https://raw.githubusercontent.com/makkus/frkl/master/tests/testfile.yaml",
        "unprocessed",
        [TEST_CONVERT_TO_PYTHON_OBJECT_DICT],
    ),
    (
        JINJA_CHAIN,
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_jinja.yaml"
        ),
        "unprocessed",
        [TESTFILE_1_CONTENT],
    ),
    (
        ABBREV_CHAIN,
        "gh:makkus/freckles/examples/quickstart.yml",
        "unprocessed",
        [
            "https://raw.githubusercontent.com/makkus/freckles/master/examples/quickstart.yml"
        ],
    ),
    (
        ABBREV_CHAIN,
        "bb:makkus/freckles/examples/quickstart.yml",
        "unprocessed",
        ["https://bitbucket.org/makkus/freckles/src/master/examples/quickstart.yml"],
    ),
    (
        FRKLIZE_CHAIN,
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_frklize_1.yml"
        ),
        "frkl",
        TEST_FRKLIZE_1_RESULT,
    ),
    (
        FRKLIZE_CHAIN,
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_frklize_2.yml"
        ),
        "frkl",
        TEST_FRKLIZE_1_RESULT,
    ),
    (
        FRKLIZE_CHAIN,
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_frklize_3.yml"
        ),
        "frkl",
        TEST_FRKLIZE_1_RESULT,
    ),
    (
        FRKLIZE_CHAIN,
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_frklize_4.yml"
        ),
        "frkl",
        TEST_FRKLIZE_1_RESULT_DOUBLE,
    ),
    (
        ABBREV_FRKLIZE_CHAIN,
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "testfile_frklize_5.yml"
        ),
        "frkl",
        TEST_FRKLIZE_1_RESULT_DOUBLE,
    ),
]


@pytest.mark.parametrize(
    "processor, input_config, context_key, expected", PROCESSOR_TESTS
)
def test_processor(processor, input_config, context_key, expected):
    frkl_obj = Frkl(input_config, processor_chain=processor)
    result = frkl_obj.process()

    pprint.pprint(result)
    print("XXX")
    pprint.pprint(expected)

    assert result == expected


@pytest.mark.parametrize("input_url", TEST_ENSURE_FAIL_URLS)
def test_ensure_fail_url_processor(input_url):
    prc = EnsureUrlProcessor()
    prc.set_current_config(input_url, {"last_call": False})
    with pytest.raises(FrklConfigException):
        prc.process()


@pytest.mark.parametrize("config, expected", [({"a": 1}, {"vars": {"a": 1}})])
def test_frkl_valid_config(config, expected):
    frkl_obj = FrklProcessor(**FRKL_INIT_PARAMS)
    frkl_obj.set_current_config(config, {"last_call": False})
    frkl_obj.process()


@pytest.mark.parametrize("config", [({"a": 1, "vars": 2}), ({"tasks": 1, "childs": 1})])
def test_frkl_invalid_config(config):
    frkl_obj = FrklProcessor(**FRKL_INIT_PARAMS)
    frkl_obj.set_current_config(config, {"last_call": False})
    with pytest.raises(FrklConfigException):
        for i in frkl_obj.process():
            print(i)


@pytest.mark.parametrize(
    "test_name",
    [
        "simple_test",
        "simple_test2",
        "simple_test3",
        "simple_test4",
        "simple_test5",
        "simple_test6",
        "simple_test7",
        "simple_test8",
        # "simple_test9",
        # "simple_test10",
        "simple_test11",
        "load_single_config_test",
        "load_multiple_configs_test",
        "download_multiple_configs_test",
        "jinja_test",
        "jinja_test_2",
        "two_input_files_1",
    ],
)
def test_files(test_name):
    folder = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "test_dirs", test_name
    )
    chain_file = os.path.join(folder, "_chain.yml")
    input_files = []
    for child in os.listdir(folder):
        if not child.startswith("_"):
            input_files.append(os.path.join(folder, child))
    input_files.sort()
    pprint.pprint(input_files)
    result_file = os.path.join(folder, "__result.yml")

    with open(result_file) as f:
        content = f.read()

    expected_obj = yaml.load(content)

    frkl_obj = factory(chain_file, input_files)
    result_obj = frkl_obj.process()

    # pprint.pprint(expected_obj)
    # print("XXX")
    pprint.pprint(result_obj)

    assert expected_obj == result_obj


@pytest.mark.parametrize("test_name", ["simple_test12"])
def test_files_collector(test_name):
    folder = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "test_dirs", test_name
    )
    chain_file = os.path.join(folder, "_chain.yml")
    input_files = []
    for child in os.listdir(folder):
        if not child.startswith("_"):
            input_files.append(os.path.join(folder, child))
    input_files.sort()
    pprint.pprint(input_files)
    result_file = os.path.join(folder, "__result.yml")

    with open(result_file) as f:
        content = f.read()

    expected_obj = yaml.load(content)

    frkl_obj = factory(chain_file, input_files)
    test12_init = {"append_keys": "vars/a"}
    result_obj = frkl_obj.process(MergeDictResultCallback(**test12_init))

    # pprint.pprint(expected_obj)
    # print("XXX")
    pprint.pprint(result_obj)

    assert expected_obj == result_obj


@pytest.mark.parametrize("test_name", ["collector_1", "collector_2", "collector_3"])
def test_collector_init(test_name):
    folder = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "test_dirs", test_name
    )

    input_files = []
    for child in os.listdir(folder):
        if not child.startswith("_"):
            input_files.append(os.path.join(folder, child))
    input_files.sort()

    result_file = os.path.join(folder, "__result.yml")

    with open(result_file) as f:
        content = f.read()

    expected_obj = yaml.load(content)

    init_file = os.path.join(folder, "_init.yml")
    result_obj = FrklCallback.init(init_file, input_files)
    result = result_obj.result()

    pprint.pprint(expected_obj)
    print("XXX")
    pprint.pprint(result)

    assert expected_obj == result



# -*- coding: utf-8 -*-
from jinja2 import Environment

FRKL_CONTEXT_DEFAULT_KEY = "frkl_vars"
ENVIRONMENT_VARS_DEFAULT_KEY = "env"

DEFAULT_LEAF_DEFAULT_KEY = "default"

STEM_KEY_NAME = "child_marker"
DEFAULT_LEAF_KEY_NAME = "default_leaf"
DEFAULT_LEAF_DEFAULT_KEY_NAME = "default_leaf_key"
OTHER_VALID_KEYS_NAME = "other_keys"
DEFAULT_LEAF_KEY_MAP_NAME = "key_move_map"

# duplication of above vars, above vars will be removed after refactoring
CHILD_MARKER_NAME = "child_marker"
DEFAULT_LEAF_NAME = "default_leaf"
DEFAULT_LEAFKEY_NAME = "default_leaf_key"
OTHER_KEYS_NAME = "other_keys"
KEY_MOVE_MAP_NAME = "key_move_map"

START_VALUES_NAME = "init_values"

ALL_FORMAT = "*"
NONE_FORMAT = "-"
PYTHON_FORMAT = "PYTHON"
STRING_FORMAT = "STRING"
DICT_FORMAT = "DICT"

FRKL_DEFAULT_PARAMS = {
    STEM_KEY_NAME: "childs",
    DEFAULT_LEAF_KEY_NAME: "task",
    DEFAULT_LEAF_DEFAULT_KEY_NAME: "task_name",
    OTHER_VALID_KEYS_NAME: ["vars"],
    DEFAULT_LEAF_KEY_MAP_NAME: "vars",
}

OPT_PLACEHOLDER = -9877
NO_STEM_INDICATOR = "-99999"
RECURSIVE_LOAD_INDICATOR = "-67323"
DEFAULT_FRKL_JINJA_ENVIRONMENT = Environment()



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import logging
import os
import sys

from frutils import StringYAML
from frutils.defaults import (
    DEFAULT_URL_ABBREVIATIONS_FILE,
    DEFAULT_URL_ABBREVIATIONS_REPO,
    JINJA_DELIMITER_PROFILES,
)
from ruamel.yaml.comments import CommentedSeq
from six import string_types
from stevedore import driver

from frkl import load_string_from_url_or_path

from .exceptions import FrklistConfigException

GLOBAL_ENV_ID_COUNTER = 1110
GLOBAL_TASKLIST_ID_COUNTER = 1110


def GLOBAL_TASKLIST_ID():
    global GLOBAL_TASKLIST_ID_COUNTER
    GLOBAL_TASKLIST_ID_COUNTER = GLOBAL_TASKLIST_ID_COUNTER + 1
    return GLOBAL_TASKLIST_ID_COUNTER


def GLOBAL_ENV_ID():
    global GLOBAL_ENV_ID_COUNTER
    GLOBAL_ENV_ID_COUNTER = GLOBAL_ENV_ID_COUNTER + 1
    return GLOBAL_ENV_ID_COUNTER


yaml = StringYAML()
yaml.default_flow_style = False

log = logging.getLogger("nsbl")


def create_tasklist(tasklist, context=None, meta=None, vars=None):
    """Loading the tasklist class

    Args:
      name (str): the registered name of the extension
      init_params (dict): the parameters to initialize the extension object

    Returns:
      DictletFinder: the extension object
    """

    log2 = logging.getLogger("stevedore")
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter("freckles plugin error -> %(message)s"))
    out_hdlr.setLevel(logging.DEBUG)
    log2.addHandler(out_hdlr)
    log2.setLevel(logging.INFO)

    if context is None:
        context = {}

    if meta is None:
        meta = {}

    if vars is None:
        vars = {}

    plugin_name = meta.get("tasklist-type", "default")

    log.debug("Loading freckles tasklist plugin '{}'...")

    mgr = driver.DriverManager(
        namespace="frkl.frkists",
        name=plugin_name,
        invoke_on_load=True,
        invoke_args=(tasklist,),
        invoke_kwds={"context": context, "meta": meta, "vars": vars},
    )

    log.debug(
        "Registered frklist: {}".format(", ".join(ext.name for ext in mgr.extensions))
    )

    return mgr.driver


class Frklist(object):
    def __init__(self, itemlist, context=None, meta=None, vars=None):

        self.itemlist_raw = itemlist
        if context is None:
            context = FrklistContext()
        self.context = context
        # if not isinstance(context, FrklistContext):
        #     self.context = copy.deepcopy(context)
        #     self.context = self.create_context(self.context)
        # else:
        #     self.context = context
        self.itemlist_raw_type = None

        if meta is None:
            meta = {}
        else:
            meta = copy.deepcopy(meta)
        self.meta = self.process_meta_properties(meta, copy.deepcopy(itemlist))

        self.itemlist_id = self.meta["list_id"]
        self.env_id = self.meta["env_id"]

        if vars is None:
            vars = {}
        else:
            vars = copy.deepcopy(vars)
        self.vars = vars

        self.itemlist_pre = self.preprocess_tasklist(
            copy.deepcopy(itemlist),
            self.context,
            copy.deepcopy(self.meta),
            copy.deepcopy(self.vars),
        )

        self.itemlist = self.expand_and_augment_tasklist(self.itemlist_pre)

    def preprocess_tasklist(self, tasklist, context, meta, vars):

        if isinstance(tasklist, string_types):
            self.itemlist_raw_type = "string"
            content = load_string_from_url_or_path(
                tasklist,
                create_python_object=True,
                template_vars=vars,
                use_environment_vars=True,
                delimiter_profile=JINJA_DELIMITER_PROFILES["freckles"],
            )
            return content
        elif isinstance(tasklist, (list, tuple, CommentedSeq)):
            self.itemlist_raw_type = "list"
            return tasklist
        else:
            raise FrklistConfigException(
                "Invalid type for tasklist: {}".format(type(tasklist))
            )

    def process_meta_properties(self, meta_dict, tasklist):

        if "list_id" not in meta_dict.keys():
            meta_dict["list_id"] = GLOBAL_TASKLIST_ID()
        if "env_id" not in meta_dict.keys():
            meta_dict["env_id"] = GLOBAL_ENV_ID()

        if isinstance(tasklist, string_types):
            meta_dict["itemlist_raw_type"] = "string"
            if "itemlist_parent" not in meta_dict.keys():
                meta_dict["itemlist_parent"] = os.path.dirname(tasklist)
        else:
            meta_dict["itemlist_raw_type"] = "list"
            if "itemlist_parent" not in meta_dict.keys():
                meta_dict["itemlist_parent"] = None

        return meta_dict

    def render_tasklist(self, **kwargs):

        return yaml.dump(self.itemlist)

    def expand_and_augment_tasklist(self, tasklist):

        return tasklist

    # def create_context(self, context_params):
    #
    #     return FrklistContext()


class DefaultFrklist(Frklist):
    def __init__(self, itemlist, **kwargs):

        super(DefaultFrklist, self).__init__(itemlist, **kwargs)


class FrklistContext(object):
    def __init__(
        self,
        base_directory=None,
        allow_remote=False,
        allow_untrusted_urls=False,
        trusted_urls=None,
        abbrevs=None,
        urls=None,
    ):
        """The context a tasklist lives in.

        This is shared among all the tasklists of the same type.

        Args:
            base_directory (str): the directory to be used for relative paths, defaults to 'os.getcwd()'
            allow_remote (bool): whether to allow the auto-download of remote artefacts (Ansible roles, tasklists, etc.
            allow_untrusted_urls (bool): if 'allow_remote', allow the download of urls that don't start with strings not contained in 'trusted_urls'
            trusted_urls (list): a list of urls (and/or abbrevs) indicating urls to download from (if 'allow_untrsuted_urls' is False
            abbrevs (dict): a dictionary of abbreviation mappings in the form of {<abbrev_alias>: <abbrev_dict>}
            urls (dict): a dictionary of paths, sorted by the type of artefacts they hold (e.g. task-aliases, Ansible roles, etc.)
        """

        if base_directory is None:
            base_directory = os.getcwd()
        self.base_directory = base_directory

        self.allow_remote = allow_remote
        self.allow_untrusted_urls = allow_untrusted_urls
        if trusted_urls is None:
            trusted_urls = []

        if self.allow_remote and not self.allow_untrusted_urls and not trusted_urls:
            log.warn(
                "'allow_untrusted_urls' is set to false, but no trusted_urls specified, setting 'allow_remote' to false"
            )
            self.allow_remote = False

        if not abbrevs:
            abbrevs = {
                "file": DEFAULT_URL_ABBREVIATIONS_FILE,
                "repo": DEFAULT_URL_ABBREVIATIONS_REPO,
            }
        self.abbrevs = abbrevs

        if urls is None:
            self.urls = {}
        else:
            self.urls = self.cleanup_environment_paths(urls)

    def cleanup_environment_paths(self, environment_paths):

        result = {}
        for path_type, paths in environment_paths.items():

            if not paths:
                paths = []
            if isinstance(paths, string_types):
                paths = [paths]
            elif isinstance(paths, tuple):
                paths = list(paths)

            paths[:] = [os.path.realpath(os.path.expanduser(p)) for p in paths]

            paths[:] = [p for p in paths if os.path.exists(p)]
            result[path_type] = paths

        return result



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

# import yaml
import io

from ruamel.yaml import YAML

from .chains import *  # noqa: F403
from .frkl import Frkl
from .processors import *  # noqa: F403

try:
    set
except NameError:
    # noinspection PyDeprecation,PyCompatibility
    from sets import Set as set


__metaclass__ = type

log = logging.getLogger("frkl")


def load_collector(name, **init_params):
    """Loading a collector extension.

    Args:
      name (str): the registered name of the collector
      init_params (dict): the parameters to initialize the extension object

    Returns:
      FrklCallback: the extension collector
    """

    log2 = logging.getLogger("stevedore")
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter("PLUGIN ERROR -> %(message)s"))
    out_hdlr.setLevel(logging.DEBUG)
    log2.addHandler(out_hdlr)
    log2.setLevel(logging.INFO)

    log.debug("Loading extension...")

    mgr = stevedore.driver.DriverManager(
        namespace="frkl.collector",
        name=name,
        invoke_on_load=True,
        invoke_kwds=init_params,
    )
    log.debug(
        "Registered plugins: {}".format(", ".join(ext.name for ext in mgr.extensions))
    )

    return mgr


# ----------------------------------------------------------------
# callbacks / collectors
@six.add_metaclass(abc.ABCMeta)
class FrklCallback(object):
    """A class to slurp up configurations from the last element of a processor chain.

    Since those processers might return Generators, it's handy to deal with the results of a config processing
    manually, callbacks seemed like a good way to do it.
    """

    def init(init_file, configs):
        """Creates a collector object with associated Frkl and processor chain.

        The init file needs to be yaml, with the 'collector' key being a registered collector plugin,
        and the 'processor_chain' key being a list of registered ConfigProcessor objects.

        Args:
          init_file: the path to the init file
          configs: the configuration item(s) for this processor
        Returns:
          FrklCallback: the collector item
        """

        yaml = YAML(typ="safe")

        with io.open(init_file, encoding="utf-8") as f:
            init_config = yaml.load(f)

        if isinstance(init_config, (list, tuple)):
            processor_chain = init_config
            collector_name = "default"
        elif not isinstance(init_config, dict):
            raise Exception(
                "init configuration needs to be either list of processor configs, or dict with 'processor_chain' and optionally 'collector' keys"
            )
        else:
            if "processor_chain" not in init_config.keys():
                raise Exception(
                    "No processor chain specified in '{}'".format(init_file)
                )

            processor_chain = init_config["processor_chain"]
            if not processor_chain:
                raise Exception("Processor chain in '{}' empty".format(init_file))

            if "collector" not in init_config.keys():
                collector_name = "default"
            else:
                collector_name = init_config["collector"]

        if isinstance(collector_name, string_types):
            collector_init = {}
        elif not isinstance(collector_name, dict) or len(collector_name) != 1:
            raise Exception(
                "'collector' value needs to be either a string or a dict with length 1"
            )
        else:
            temp = collector_name
            collector_name = list(temp.keys())[0]
            collector_init = temp[collector_name]

        if collector_name == "default":
            collector_name = "merge"

        collector = load_collector(collector_name, **collector_init).driver
        bootstrap = Frkl(processor_chain, COLLECTOR_INIT_BOOTSTRAP_PROCESSOR_CHAIN)
        config_frkl = bootstrap.process(FrklFactoryCallback())

        config_frkl.set_configs(configs)

        temp = config_frkl.process(collector)
        return collector

    init = staticmethod(init)

    def __init__(self, **init_params):

        self.init_params = init_params

    @abc.abstractmethod
    def callback(self, item):
        """Adds a new item to the callback class.

        Args:
          item (object): the newly processed config
        """
        pass

    @abc.abstractmethod
    def result(self):
        """Returns a meaningful representation of all added configs so far.

        Ideally this is a string representation of the (current) state of the callback, since it might
        be used by 3rd party tools for debugging purposes, or as a method to show state in the absence
        of knowledge of the type of FrklCallback that is used.

        Returns:
          object: the current state of the callback
        """
        pass

    def started(self):
        """Optional method that can be overwritten if the callback needs to know when the processing has started."""

        pass

    def finished(self):
        """Optional method that can be overwritten if the callback needs to know when the processing has finished."""

        pass


class MergeDictResultCallback(FrklCallback):
    """Simple callback, merges all configs to *one* internal dict."""

    def __init__(self, **init_params):

        super(MergeDictResultCallback, self).__init__(**init_params)

        self.result_dict = {}
        self.append_keys = init_params.get("append_keys", [])
        if not isinstance(self.append_keys, (list, tuple)):
            self.append_keys = [self.append_keys]
        self.append_keys_map = {}

    def get_dict_detail(self, target_dict, detail_path):

        temp = target_dict
        for level in detail_path.split("/"):
            temp = temp.get(level, {})

        return temp

    def set_dict_detail(self, target_dict, detail_path, new_value):

        temp = target_dict
        tokens = detail_path.split("/")
        for level in tokens[0:-1]:
            if level not in temp.keys():
                temp[level] = {}
            temp = temp[level]

        temp[tokens[-1]] = new_value

    def callback(self, process_result):

        for ak in self.append_keys:
            v = self.get_dict_detail(process_result, ak)
            if v:
                if not isinstance(v, (list, tuple)):
                    v = [v]
                self.append_keys_map.setdefault(ak, []).extend(v)

        dict_merge(self.result_dict, process_result, copy_dct=False)

    def result(self):

        for k, v in self.append_keys_map.items():
            self.set_dict_detail(self.result_dict, k, v)

        return self.result_dict


class SetResultCallback(FrklCallback):
    """Simple callback to create a set out of a list."""

    def __init__(self, **init_params):

        super(SetResultCallback, self).__init__(**init_params)

        self.result_set = set()
        self.return_list = init_params.get("return_list", False)

    def callback(self, process_result):

        self.result_set.add(process_result)

    def result(self):

        if self.return_list:
            return list(self.result_set)
        else:
            return self.result_set


class MergeResultCallback(FrklCallback):
    """Simple callback, just appends all configs to an internal list."""

    def __init__(self, **init_params):
        super(MergeResultCallback, self).__init__(**init_params)

        self.result_list = []

    def callback(self, process_result):
        self.result_list.append(process_result)

    def result(self):
        return self.result_list


class ExtendResultCallback(FrklCallback):
    """Simple callback, extends an internal list with the processing results.
    """

    def __init__(self, **init_params):
        super(ExtendResultCallback, self).__init__(**init_params)
        self.result_list = []

    def callback(self, process_result):
        self.result_list.extend(process_result)

    def result(self):
        return self.result_list


class FrklFactoryCallback(FrklCallback):
    """Helper callback method, creates a new Frkl object by processing a list of processor init dicts.
    """

    def __init__(self, **init_params):
        super(FrklFactoryCallback, self).__init__(**init_params)
        self.processors = []
        self.bootstrap_chain = []

    def callback(self, item):
        self.processors.append(item)

        ext_name = item.get("processor", {}).get("type", None)
        if not ext_name:
            raise FrklConfigException(
                "Can't parse processor name using config: {}".format(item)
            )
        ext_init_params = item.get("init", {})
        log.debug(
            "Loading extension '{}' using init parameters: '{}".format(
                ext_name, ext_init_params
            )
        )
        ext = load_extension(ext_name, **ext_init_params)
        self.bootstrap_chain.append(ext.driver)

    def result(self):
        return Frkl([], self.bootstrap_chain)



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

from .chains import *  # noqa: F403
from .processors import *  # noqa: F403

__metaclass__ = type

log = logging.getLogger("frkl")


class Frkl(object):
    def __init__(
        self, configs=None, processor_chain=LOAD_OBJECT_FROM_URL_CHAIN, max_items=9999
    ):
        """Base object that holds the configuration.

        Args:
          configs (list): list of configurations, will be processed in the order they come in
          processor_chain (list): processor chain to use, defaults to [:class:`UrlAbbrevProcessor`]
          max_items (int): max number of items to allow, mostly used to catch bugs in the code, contact the developer if you hit this number
L        """

        if configs is None:
            configs = []
        if not isinstance(processor_chain, (list, tuple)):
            processor_chain = [processor_chain]
        self.processor_chain = processor_chain

        self.configs = []
        self.set_configs(configs)
        self.max_items = max_items

    def set_configs(self, configs):
        """Sets the configuration(s) for this Frkl object.

        Args:
          configs (list): the configurations, will wrapped in a list if not a list or tuple already
        """

        if not isinstance(configs, (list, tuple)):
            configs = [configs]

        self.configs = list(configs)

    def append_configs(self, configs):
        """Appends the provided configuration(s) for this Frkl object.

        Args:
          configs (list): the configurations, will wrapped in a list if not a list or tuple already
        """

        if not isinstance(configs, (list, tuple)):
            configs = [configs]

        # ensure configs are wrapped
        for c in configs:
            self.configs.append(c)

    def process(self, callback=None):
        """Kicks off the processing of the configuration urls.

      Args:
        callback (FrklCallback): callback to use for this processing run, defaults to 'MergeResultCallback'

      Returns:
        object: the value of the result() method of the callback
      """

        if not callback:
            from .callbacks import MergeResultCallback

            callback = MergeResultCallback()

        configs_copy = copy.deepcopy(self.configs)
        context = {"last_call": False}

        callback.started()

        while configs_copy:

            if len(configs_copy) > self.max_items:
                raise FrklConfigException(
                    "More than {} configs, this looks like a loop, exiting. Contact the developer to up that limit if you think this is an issue.".format(
                        self.max_items
                    )
                )

            config = configs_copy.pop(0)
            context["current_original_config"] = config

            self.process_single_config(
                config, self.processor_chain, callback, configs_copy, context
            )

        current_config = None
        context["next_configs"] = []

        context["current_config"] = current_config
        context["last_call"] = True
        self.process_single_config(
            current_config, self.processor_chain, callback, [], context
        )

        callback.finished()

        return callback.result()

    def process_single_config(
        self, config, processor_chain, callback, configs_copy, context
    ):
        """Helper method to be able to recursively call the next processor in the chain.

        Args:
          config (object): the current config object
          processor_chain (list): the list of processor items to use (reduces by one with every recursive run)
          callback (FrklCallback): the callback that receives any potential results
          configs_copy (list): list of configs that still need processing, this method might prepend newly processed configs to this
          context (dict): context object, can be used by processors to investigate current state, history, etc.
        """

        if not context.get("last_call", False):
            if not config:
                return

        if not processor_chain:
            if config:
                callback.callback(config)
            return

        current_processor = processor_chain[0]
        # TODO: check if this needs deepcopy
        temp_config = copy.copy(config)

        context["current_processor"] = current_processor
        context["current_config"] = temp_config
        context["current_processor_chain"] = processor_chain
        context["next_configs"] = configs_copy

        current_processor.set_current_config(temp_config, context)

        additional_configs = current_processor.get_additional_configs()
        if additional_configs:
            configs_copy[0:0] = additional_configs

        last_processing_result = current_processor.process()
        if isinstance(last_processing_result, types.GeneratorType):
            for item in last_processing_result:
                self.process_single_config(
                    item, processor_chain[1:], callback, configs_copy, context
                )

        else:
            self.process_single_config(
                last_processing_result,
                processor_chain[1:],
                callback,
                configs_copy,
                context,
            )


# several useful or common helper methods


def load_object_from_url_or_path(urls):
    """Simple wrapper to create a list of dictionaries from a local or remote file.

    If input is a single url, a single list will be returned. If a list,
    the result will be a list of lists.

    As this uses safe_load with the yaml parser, the dictionary will probably not be in the same order as in the original file.

    Args:
        urls (list): a list of paths and/or urls
        safe_load (bool): whether to use safe load with the yaml parser. This will destroy the order of a dict.
    Returns:
        list: a list of dictionaries, representing the content of the input files
    """
    single_input = False
    if isinstance(urls, string_types):
        single_input = True
        urls = [urls]

    f = Frkl(urls, LOAD_OBJECT_FROM_URL_CHAIN)
    result = f.process()

    if single_input:
        return result[0]
    else:
        return result


def load_string_from_url_or_path(
    urls,
    template_vars=None,
    delimiter_profile=JINJA_DELIMITER_PROFILES["default"],
    use_environment_vars=False,
    use_context=False,
    create_python_object=False,
    safe_load=True,
):
    """Simple wrapper to create a list of dictionaries from a local or remote file.

    If input is a single url, a single list will be returned. If a list,
    the result will be a list of lists.

    Args:
        urls (list): a list of paths and/or urls
        template_vars (dict): if not None, the string will be considered a jinja2 template and this dict will be used as replacemnt dict
        delimiter_profile (dict): the delimiter profile, if templating
        use_environment_vars (bool, str): whether to also use environment variables when templating
        use_context (bool, str): whether to also use the frkl context when templating
        create_python_object (bool): whether to create a python object out of the result string or not
    Returns:
        list: a string or python object, representing the content of the input files
    """

    single_input = False
    if isinstance(urls, string_types):
        single_input = True
        urls = [urls]

    chain = load_templated_string_from_url_chain(
        template_vars,
        create_python_object=create_python_object,
        use_environment_vars=use_environment_vars,
        use_context=use_context,
        delimiter_profile=delimiter_profile,
    )

    f = Frkl(urls, chain)
    result = f.process()

    if single_input:
        if not result:
            return []
        return result[0]
    else:
        return result



# -*- coding: utf-8 -*-
import json
import logging
from collections import Mapping

import click
from omegaconf import OmegaConf, DictConfig

from frutils import StringYAML, dict_merge, is_url_or_abbrev
from frutils.defaults import *  # noqa: F403
from ruamel.yaml.comments import CommentedSeq
from six import string_types

from .callbacks import (
    EnsurePythonObjectProcessor,
    EnsureUrlProcessor,
    SetResultCallback,
)
from .frkl import Frkl
from .processors import ParentPathProcessor, UrlAbbrevProcessor

log = logging.getLogger("frkl")

yaml = StringYAML()


def expand_string_to_git_details(value, default_abbrevs):

    branch = None
    opt_split_string = "::"
    if opt_split_string in value:
        tokens = value.split(opt_split_string)
        opt = tokens[1:-1]
        if not opt:
            raise Exception(
                "Not a valid url, needs at least 2 split strings ('{}')".format(
                    opt_split_string
                )
            )
        if len(opt) != 1:
            raise Exception("Not a valid url, can only have 1 branch: {}".format(value))
        branch = opt[0]

    result = expand_string_to_git_repo(value, default_abbrevs)
    result = {"url": result}

    if branch:
        result["branch"] = branch

    return result


def expand_string_to_git_repo(value, default_abbrevs):
    if isinstance(value, string_types):
        is_string = True
    elif isinstance(value, (list, tuple)):
        is_string = False
    else:
        raise Exception(
            "Not a supported type (only string or list are accepted): {}".format(value)
        )

    try:
        init_params = {"abbrevs": default_abbrevs, "add_default_abbrevs": False}
        frkl_obj = Frkl(value, [UrlAbbrevProcessor(**init_params)])
        result = frkl_obj.process()
        if is_string:
            return result[0]
        else:
            return result
    except (Exception) as e:
        raise Exception("'{}' is not a valid repo url: {}".format(value, e))


def get_url_parents(urls, abbrevs=False, return_list=False):
    """Helper methods to calculate the parents of the provided urls.

    Args:
        urls (list): the list of urls
        abbrevs (bool, dict): if False, urls won't be expanded if they are abbreviated, otherwise if a abbrev dict is provided, they will be
        return_list (bool): whether to return the result as set (False) or list (True)
    """

    if abbrevs is False or abbrevs is None:
        chain = [ParentPathProcessor()]
    else:
        if abbrevs is True:
            abbrevs = DEFAULT_URL_ABBREVIATIONS_FILE
        chain = [
            UrlAbbrevProcessor(
                init_params={"abbrevs": abbrevs, "add_default_abbrevs": False}
            ),
            ParentPathProcessor(),
        ]
    callback = SetResultCallback(init_params={"return_list": return_list})

    frkl_obj = Frkl(urls, chain)
    result = frkl_obj.process(callback)

    return result


class VarsFileType(click.ParamType):

    name = "vars_file_type"

    def convert(self, value, param, ctx):

        chain = [
            UrlAbbrevProcessor(),
            EnsureUrlProcessor(),
            EnsurePythonObjectProcessor(),
            # LoadMoreConfigsProcessor(),
        ]

        try:
            if not isinstance(value, (list, tuple, CommentedSeq)):
                value = [value]

            frkl_obj = Frkl(value, chain)
            result = frkl_obj.process()

            if isinstance(result[0], (list, tuple)):

                result_dict = {}
                for item in result[0]:
                    dict_merge(result_dict, item, copy_dct=False)

                return result_dict
            else:
                return result[0]

        except (Exception) as e:
            self.fail("Can't read vars '{}': {}".format(value, str(e)))


class VarsType(click.ParamType):

    name = "vars_type"

    def convert(self, value, param, ctx):

        value = self._convert(value, param, ctx)
        if isinstance(value, DictConfig):
            return value
        else:
            oc = OmegaConf.create(value)
        return oc

    def _convert(self, value, param, ctx):

        path = os.path.realpath(os.path.expanduser(value))
        value_type = "string"

        if os.path.exists(path):
            value_type = "local"
        elif is_url_or_abbrev(value):
            value_type = "remote"

        if value_type in ["remote", "local"]:
            chain = [
                UrlAbbrevProcessor(),
                EnsureUrlProcessor(),
                EnsurePythonObjectProcessor(),
                # LoadMoreConfigsProcessor(),
            ]

            try:
                if not isinstance(value, (list, tuple, CommentedSeq)):
                    value = [value]

                frkl_obj = Frkl(value, chain)
                result = frkl_obj.process()

                if isinstance(result[0], (list, tuple)):

                    result_dict = {}
                    for item in result[0]:
                        dict_merge(result_dict, item, copy_dct=False)

                    return result_dict
                else:
                    return result[0]

            except (Exception) as e:
                self.fail("Can't read vars url '{}': {}".format(value, str(e)))

        else:

            # try json
            log.debug("Trying json for value: '{}'".format(value))
            try:
                result = json.loads(value)
                if not isinstance(result, Mapping):
                    self.fail("Can't parse variable string as json: {}".format(result))
            except (Exception):
                log.debug("Trying yaml for value: '{}'".format(value))
                try:
                    result = yaml.load(value)
                    if not isinstance(result, Mapping):
                        self.fail(
                            "Can't parse variable string as yaml: {}".format(result)
                        )
                except (Exception):
                    if "=" in value:
                        return OmegaConf.from_dotlist([value])
                    else:
                        self.fail("Can't read vars string: {}".format(value))

            return result



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import abc
import io
import logging
import os
import sys
import types
from collections import Sequence

import requests
import six
import stevedore
from frutils.frutils import *  # noqa: F403
from six import string_types

from .defaults import *  # noqa: F403
from .exceptions import FrklConfigException

yaml = StringYAML(typ="safe")
yaml.default_flow_style = False

try:
    set
except NameError:
    # noinspection PyDeprecation,PyCompatibility
    from sets import Set as set

__metaclass__ = type

log = logging.getLogger("frkl")


# ------------------------------------------------------------
# utility methods


# extensions
# ------------------------------------------------------------------------
def load_extension(name, **init_params):
    """Loading a processor extension.

    Args:
      name (str): the registered name of the extension
      init_params (dict): the parameters to initialize the extension object

    Returns:
      FrklConfig: the extension object
    """

    log2 = logging.getLogger("stevedore")
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(logging.Formatter("PLUGIN ERROR -> %(message)s"))
    out_hdlr.setLevel(logging.DEBUG)
    log2.addHandler(out_hdlr)
    log2.setLevel(logging.INFO)

    log.debug("Loading extension...")

    print(name)
    print(init_params)

    mgr = stevedore.driver.DriverManager(
        namespace="frkl.frk", name=name, invoke_on_load=True, invoke_kwds=init_params
    )

    log.debug(
        "Registered plugins: {}".format(", ".join(ext.name for ext in mgr.extensions))
    )

    return mgr


@six.add_metaclass(abc.ABCMeta)
class ConfigProcessor(object):
    """Abstract base class for config url/content manipulators.

    In order to enable configuration urls and content to be written as quickly and minimal as possible, frkl supports pluggable processors that can manipulate the configuration urls and contents. For example, urls can be abbreviated 'gh' -> 'https://raw.githubusercontent.com/blahblah'.
    """

    def __init__(self, **init_params):
        """
        Args:
          init_params (dict): arguments to initialize the processor
        """

        self.init_params = init_params

        self.current_input_config = None
        self.current_context = None
        self.last_call = False

    def get_input_format(self):
        """Returns the format of the accepted input.

        Defaults to 'STRING'
        """

        return STRING_FORMAT

    def get_output_format(self):
        """Returns the format of the output.

        Defaults to the same as the 'get_input_format' method.
        """

        return self.get_input_format()

    def set_current_config(self, input_config, context):
        """Sets the current configuration.

        Calls the 'new_config' method after assigning the new input configuration to the 'self.current_input_config' variable.

        Args:
          input_config (object): current configuration to be processed
          context (dict): dict that describes the current context / processing state
        """

        if isinstance(input_config, types.GeneratorType):
            raise Exception("Can't deal with Type 'Generator' as a value.")

        self.current_input_config = input_config
        self.current_context = context

        self.last_call = self.current_context["last_call"]
        self.new_config()

    def new_config(self):
        """Can be overwritten to initially compute a new config after it is first set.

        The newly (last) added input configuration is stored in the 'self.current_input_config' variable.
        """

        pass

    def get_additional_configs(self):
        """Returns additional configs if applicable.

        This is called before the 'process' method.

        Returns:
          list: other configs to be processed next
        """

        return None

    def process(self):
        """Processes the config url or content.

        Calls the 'process_current_config' method internally

        Returns:
          object: processed config url or content
        """

        if self.last_call:
            if self.current_input_config or self.handles_last_call():
                return self.process_current_config()
            else:
                return None
        else:
            return self.process_current_config()

    def handles_last_call(self):
        """Returns whether this processor wants to be called at the end of a processing run again.

        If the preceding processor returns a non-None value, this is ignored and the processor is called anyway.

        Returns:
          bool: whether to call this processor for the 'special' last run
        """

        return False

    @abc.abstractmethod
    def process_current_config(self):
        """Processes the config url or content.

        Returns:
          object: processed config url or content
        """

        pass


class EnsureUrlProcessor(ConfigProcessor):
    """Makes sure the provided string is a url, then downloads the target and reads the content."""

    def get_config(self, config_file_url):
        """Retrieves the config (if necessary), and returns its content.

        Config can be either a path to a local yaml file, an url to a remote yaml file, or a json string.

        Args:
          config_file_url (str): the url/path/json content
        """

        # check if file first
        if os.path.exists(config_file_url):
            log.debug("Opening as file: {}".format(config_file_url))
            with io.open(config_file_url, encoding="utf-8") as f:
                content = f.read()

        # check if url
        elif config_file_url.startswith("http"):

            log.debug("Opening as url: {}".format(config_file_url))
            verify_ssl = True
            try:
                r = requests.get(config_file_url, verify=verify_ssl)
                r.raise_for_status()
                content = r.text
            except (Exception) as e:
                raise FrklConfigException(
                    "Could not retrieve url: {}".format(config_file_url), e
                )
        else:
            raise FrklConfigException(
                "Not a supported config file url or no local file found: {}".format(
                    config_file_url
                )
            )

        return content

    def process_current_config(self):

        result = self.get_config(self.current_input_config)
        return result


class ParentPathProcessor(ConfigProcessor):
    def get_output_format(self):
        return STRING_FORMAT

    def get_input_format(self):
        return STRING_FORMAT

    def process_current_config(self):

        url = self.current_input_config

        if not is_url_or_abbrev(url) and not os.path.isabs((url)):
            url = os.path.abspath(url)

        parent = os.path.dirname(url)

        return parent


class RemoveKeysProcessor(ConfigProcessor):
    def __init__(self, keys_to_remove=None, **kwargs):

        super(RemoveKeysProcessor, self).__init__(**kwargs)
        if keys_to_remove is None:
            keys_to_remove = []
        self.keys_to_remove = keys_to_remove

    def get_output_format(self):
        return DICT_FORMAT

    def get_input_format(self):
        return DICT_FORMAT

    def process_current_config(self):

        new_config = self.current_input_config

        if isinstance(new_config, (list, tuple, CommentedSeq)):
            for c in new_config:
                if isinstance(c, (dict, CommentedMap, OrderedDict)):
                    for key in self.keys_to_remove:
                        c.pop(key, None)

        elif isinstance(new_config, (dict, CommentedMap, OrderedDict)):
            for key in self.keys_to_remove:
                new_config.pop(key, None)

        return new_config


class EnsurePythonObjectProcessor(ConfigProcessor):
    """Makes sure the provided string is either valid yaml (or json -- not implemented yet), and converts it into a python object.

    Args:
        safe (bool): whether to use the 'safe' type for the yaml parser. This will destroy the order of any source dicts.
    """

    def __init__(self, safe_load=True, **kwargs):

        super(EnsurePythonObjectProcessor, self).__init__(**kwargs)
        self.safe_load = safe_load

    def get_output_format(self):
        return "PYTHON"

    def process_current_config(self):

        if not self.safe_load:
            temp = StringYAML()
            config_obj = temp.load(self.current_input_config)
        else:
            config_obj = yaml.load(self.current_input_config)
        return config_obj


class ToYamlProcessor(ConfigProcessor):
    """Takes a python object and returns the string representation.
    """

    def get_input_format(self):
        return PYTHON_FORMAT

    def get_output_format(self):
        return STRING_FORMAT

    def process_current_config(self):
        result = yaml.dump(self.current_input_config)
        return result


class IdProcessor(ConfigProcessor):
    """Adds an id to every config item."""

    def __init__(self, **init_params):

        super(IdProcessor, self).__init__(**init_params)
        self.id_type = init_params.get("id_type", "enumerate")
        self.id_name = init_params.get("id_name", "id")
        self.id_key = init_params.get("id_key", False)

        if not self.id_key:
            raise FrklConfigException("No 'id_key' value provided for IdProcessor")

        self.current_id = 0

    def get_input_format(self):
        return PYTHON_FORMAT

    def validate_init(self):

        return True

    def process_current_config(self):
        self.current_input_config[self.id_key][self.id_name] = self.current_id
        self.current_id = self.current_id + 1

        return self.current_input_config


class MergeProcessor(ConfigProcessor):
    """Gathers all configs and returns a list of all results as single element."""

    def __init__(self, **init_params):

        super(MergeProcessor, self).__init__(**init_params)

    def get_input_format(self):

        return ALL_FORMAT

    def new_config(self):
        if not self.last_call:
            self.configs.append(self.current_input_config)

    def process_current_config(self):

        if self.last_call:
            return self.configs
        else:
            return None

    def handles_last_call(self):
        return True


class DictInjectionProcessor(ConfigProcessor):
    """A processor to 'inject' dictionaries and dictionary values into other dictionaries, according to predefined rules.
    """

    def __init__(self, **init_params):

        super(DictInjectionProcessor, self).__init__(**init_params)
        self.injection_dicts = init_params["injection_dicts"]
        if isinstance(self.injection_dicts, dict):
            self.injection_dicts = [self.injection_dicts]

        self.on_top = init_params.get("merge_on_top", False)
        self.separator = init_params.get("key_separator", "/")

    def get_input_format(self):

        return PYTHON_FORMAT

    def process_current_config(self):

        config = self.current_input_config

        result = None
        for inj_dict in self.injection_dicts:

            for key, value in inj_dict.items():

                key_hierarchy = key.split(self.separator)
                current_config = config
                for part_key in key_hierarchy:

                    if part_key in current_config.keys():
                        current_config = current_config[part_key]
                    else:
                        current_config = None
                        break

                if not current_config or current_config not in value.keys():
                    continue

                merge_dict = inj_dict[key][current_config]

                if self.on_top:
                    result = dict_merge(config, merge_dict)
                else:
                    result = dict_merge(merge_dict, config)

                config = result

        return result


class FrklProcessor(ConfigProcessor):
    """A processor to 'expand' python dictionaries using a pre-defined schema.

    This is a bit more complicated to explain than I'd like it to be. For that reason, there is an extra
    page in the docs: link (XXX)
    """

    def __init__(self, **init_params):

        # self.stem_key = None
        # self.default_leaf_key = None
        # self.default_leaf_default_key = None
        # self.other_valid_keys = None
        # self.default_leaf_key_map = None
        # self.all_keys = None
        # self.use_context = None
        # self.values_so_far = None

        super(FrklProcessor, self).__init__(**init_params)

        self.stem_key = init_params[STEM_KEY_NAME]
        self.default_leaf_key = init_params[DEFAULT_LEAF_KEY_NAME]
        self.default_leaf_default_key = init_params.get(
            DEFAULT_LEAF_DEFAULT_KEY_NAME, None
        )
        if self.default_leaf_default_key is None:
            raise FrklConfigException("No default leaf default key specified.")
        self.other_valid_keys = init_params.get(OTHER_VALID_KEYS_NAME, [])
        self.default_leaf_key_map = init_params.get(DEFAULT_LEAF_KEY_MAP_NAME, None)
        if self.default_leaf_key_map is None:
            self.default_leaf_key_map = self.default_leaf_key

        if isinstance(self.default_leaf_key_map, string_types):
            if "/" in self.default_leaf_key_map:
                tokens = self.default_leaf_key_map.split("/")
                if not len(tokens) == 2:
                    raise FrklConfigException(
                        "Default value for move_key_map can't be parsed as it has more than 2 parts (separated by '/': {})".format(
                            self.default_leaf_key_map
                        )
                    )
                self.default_leaf_key_map = {"*": (tokens[0], tokens[1])}
            else:
                self.default_leaf_key_map = {
                    "*": (self.default_leaf_key_map, DEFAULT_LEAF_DEFAULT_KEY)
                }
        elif isinstance(self.default_leaf_key_map, dict):
            # make sure the move_key_map has the right format
            for key in self.default_leaf_key_map.keys():
                value = self.default_leaf_key_map[key]
                if isinstance(value, (list, tuple)):
                    if not len(value) == 2:
                        raise FrklConfigException(
                            "Value for move_key_map can't be parsed as it has more than 2 parts (separated by '/': {})".format(
                                value
                            )
                        )
                    self.default_leaf_key_map[key] = value
                else:
                    if not isinstance(value, string_types) and not len(value) == 2:
                        raise FrklConfigException(
                            "move_key_map needs a list or tuple as value type with length '2': {}".format(
                                self.default_leaf_key_map
                            )
                        )

                    if "/" in value:
                        tokens = value.split("/")

                        if not len(tokens) == 2:
                            raise FrklConfigException(
                                "Value for move_key_map can't be parsed as it has more than 2 parts (separated by '/': {})".format(
                                    value
                                )
                            )
                        self.default_leaf_key_map[key] = (tokens[0], tokens[1])
                    else:
                        self.default_leaf_key_map[key] = (
                            value,
                            DEFAULT_LEAF_DEFAULT_KEY,
                        )

        else:
            raise FrklConfigException(
                "Type '{}' not supported for move_key_map.".format(
                    type(self.default_leaf_key_map)
                )
            )

        self.all_keys = set([self.stem_key, self.default_leaf_key])
        self.all_keys.update(self.other_valid_keys)

        for item in self.default_leaf_key_map.values():
            self.all_keys.add(item[0])

        self.use_context = init_params.get("use_context", False)
        if self.use_context and isinstance(self.use_context, bool):
            self.use_context = FRKL_CONTEXT_DEFAULT_KEY
        elif self.use_context and not isinstance(self.use_context, string_types):
            raise FrklConfigException(
                "'use_context' keyword needs to be of type bool or string: {}".format(
                    init_params
                )
            )

        if START_VALUES_NAME in init_params.keys():
            self.values_so_far = init_params[START_VALUES_NAME]
        else:
            self.values_so_far = {}

        self.configs = []

    def get_input_format(self):

        return PYTHON_FORMAT

    def new_config(self):

        # make sure the new value is a dict, with only allowed keys
        if isinstance(self.current_input_config, (list, tuple)):
            self.configs.extend(self.current_input_config)
        else:
            self.configs.append(self.current_input_config)

        if self.use_context:
            self.current_context[self.use_context] = self.values_so_far

    def process_current_config(self):

        result = self.frklize(self.current_input_config, self.values_so_far)
        return result

    def frklize(self, config, current_vars):
        """Recursively called function which generates (expands) and yields dictionaries matching
        certain criteria (containing leaf_node keys, for example).

        Args:
          config (object): the input config
          current_vars (dict): current state of the (overlayed) var cache
        """

        # making sure the new value is a dict, with only allowed keys
        if isinstance(config, string_types):
            config = {self.default_leaf_key: {self.default_leaf_default_key: config}}

        if isinstance(config, (list, tuple, Sequence)):
            for item in config:
                for result in self.frklize(item, copy.deepcopy(current_vars)):
                    yield result
        else:

            if not isinstance(config, (dict, CommentedMap, OrderedDict)):

                raise FrklConfigException(
                    "Not a supported type for value '{}': {}".format(
                        config, type(config)
                    )
                )

            new_value = {}

            # check whether any of the known keys is available here, if not,
            # we check whether there is a default key registered for the name of the keys
            if not any(x in config.keys() for x in self.all_keys):

                if not len(config) == 1:
                    raise FrklConfigException(
                        "This form of configuration is not implemented yet (can't have more than one key if key is not among the known keys): {} -- current vars: {}".format(
                            config, current_vars
                        )
                    )
                else:
                    key = next(iter(config))
                    value = config[key]

                    insert_leaf_key = self.default_leaf_key
                    insert_leaf_key_key = self.default_leaf_default_key
                    new_value.setdefault(insert_leaf_key, {})[insert_leaf_key_key] = key

                    if not isinstance(value, dict):
                        if key in self.default_leaf_key_map.keys():
                            val_key = self.default_leaf_key_map[key][0]
                            val_key_key = self.default_leaf_key_map[key][1]
                        elif "*" in self.default_leaf_key_map.keys():
                            val_key = self.default_leaf_key_map["*"][0]
                            val_key_key = self.default_leaf_key_map["*"][1]
                        else:
                            raise FrklConfigException(
                                "Can't find entry in move_key_map for key '{}' in order to move value: {}"
                            ).format(key, value)
                        new_value.setdefault(val_key, {})[val_key_key] = value

                    else:
                        if all(x in self.all_keys for x in value.keys()):
                            dict_merge(new_value, value, copy_dct=False)
                        elif all(x not in self.all_keys for x in value.keys()):
                            if key in self.default_leaf_key_map.keys():
                                migrate_key = self.default_leaf_key_map[key][0]
                            elif "*" in self.default_leaf_key_map.keys():
                                migrate_key = self.default_leaf_key_map["*"][0]
                            else:
                                raise FrklConfigException(
                                    "Can't find default_leaf_key to move values of key '{}".format(
                                        key
                                    )
                                )

                            new_value.setdefault(migrate_key, {}).update(value)
                            new_value[migrate_key].update(value)
                        else:
                            for key, value in value.items():

                                if key in self.all_keys:
                                    if isinstance(value, dict):
                                        dict_merge(
                                            new_value.setdefault(key, {}),
                                            value,
                                            copy_dct=False,
                                        )
                                    else:
                                        new_value[key] = value

                                else:
                                    if key in self.default_leaf_key_map.keys():
                                        migrate_key = self.default_leaf_key_map[key][0]
                                    elif "*" in self.default_leaf_key_map.keys():
                                        migrate_key = self.default_leaf_key_map["*"][0]
                                    else:
                                        raise FrklConfigException(
                                            "Can't find default_leaf_key to move values of key '{}".format(
                                                key
                                            )
                                        )

                                    if isinstance(value, dict):
                                        new_value.setdefault(migrate_key, {}).update(
                                            value
                                        )
                                        new_value[migrate_key].update(value)
                                    else:
                                        new_value.setdefault(migrate_key, {})[
                                            key
                                        ] = value

                            # raise FrklConfigException("Mixed keys.")

            else:
                # check whether all keys are allowed

                for key in config.keys():
                    if key not in self.all_keys:
                        raise FrklConfigException(
                            "Key '{}' not allowed, since it is an unknown keys amongst known keys in config: {}".format(
                                key, config
                            )
                        )

                new_value = config

                # if self.stem_key in new_value.keys() and self.default_leaf_key in new_value.keys():
                # raise FrklConfigException(
                # "Configuration can't have both stem key ({}) and default leaf key ({}) on the same level: {}".
                # format(self.stem_key, self.default_leaf_key, new_value))

            # at this point we have an 'expanded' dict

            stem_branch = new_value.pop(self.stem_key, NO_STEM_INDICATOR)
            # merge new values with current_vars
            dict_merge(current_vars, new_value, copy_dct=False)
            # TODO: check if this needs deepcopy
            new_value = copy.copy(current_vars)

            if stem_branch == NO_STEM_INDICATOR:
                # TODO: double check logic here
                # if self.default_leaf_key in new_value.keys() and self.default_leaf_default_key in new_value[self.default_leaf_key].keys():
                if self.default_leaf_key in new_value.keys():
                    yield new_value

            else:
                # TODO: check if this needs deepcopy
                for item in self.frklize(stem_branch, copy.copy(current_vars)):
                    yield item


class Jinja2TemplateProcessor(ConfigProcessor):
    """Processor to replace all occurrences of Jinja template strings with values (predefined,
    or potentially dynamically processed in an earlier step).

    Args:
        template_values (dict): a dictionary containing the values to replace template strings with
        use_environment_vars (bool): whether to also use current environment variables  in the replacement dict. If True, it'll stored under the key 'LOCAL_ENV', if string, that will be used as key
        use_context (bool): whether to also use the frkl context in the replacement dict.  If True, it'll stored under the key 'frkl_vars', if string, that will be used as key
        delimiter_profile (dict): the jinja2 delimters
        **init_params (dict): additional config for the parent class
    """

    def __init__(
        self,
        template_values=None,
        use_environment_vars=False,
        use_context=False,
        delimiter_profile=JINJA_DELIMITER_PROFILES["default"],
        **init_params
    ):

        super(Jinja2TemplateProcessor, self).__init__(**init_params)
        if not template_values:
            template_values = {}
        self.template_values = template_values
        self.use_environment_vars = use_environment_vars
        if self.use_environment_vars and isinstance(self.use_environment_vars, bool):
            self.use_environment_vars = DEFAULT_LOCAL_ENV_VARS_KEY
        elif self.use_environment_vars and not isinstance(
            self.use_environment_vars, string_types
        ):
            raise FrklConfigException(
                "'use_environment_vars' keyword needs to be of type bool or string: {}".format(
                    use_environment_vars
                )
            )

        self.use_context = use_context
        if self.use_context and isinstance(self.use_context, bool):
            self.use_context = FRKL_CONTEXT_DEFAULT_KEY
        elif self.use_context and not isinstance(self.use_context, string_types):
            raise FrklConfigException(
                "'use_context' keyword needs to be of type bool or string: {}".format(
                    use_context
                )
            )
        self.jinja_env = Environment(**delimiter_profile)

    def get_input_format(self):

        return STRING_FORMAT

    def process_current_config(self):

        old_string = self.current_input_config

        env = {}

        if self.use_context:
            frkl_vars = self.current_context.get(self.use_context, {})
            dict_merge(env, frkl_vars, copy_dct=False)

        dict_merge(env, self.template_values, copy_dct=False)

        new_string = replace_string(
            old_string,
            replacement_dict=env,
            jinja_env=self.jinja_env,
            local_env_vars_key=self.use_environment_vars,
        )
        return new_string


class YamlTextSplitProcessor(ConfigProcessor):
    """Splits a string if it can find certain keywords at the beginning of a line.

    Args:
        keywords (list): a list of keywords
    """

    def __init__(self, **init_params):

        super(YamlTextSplitProcessor, self).__init__(**init_params)
        self.keywords = init_params["keywords"]
        self.current_lines = []

    def get_input_format(self):
        return STRING_FORMAT

    def process_current_config(self):

        if self.last_call:
            if self.current_lines:
                yield "\n".join(self.current_lines)
        else:
            new_config = self.current_input_config

            for line in new_config.splitlines():
                if self.current_lines and any(
                    line.startswith(keyword) for keyword in self.keywords
                ):
                    yield "\n".join(self.current_lines)
                    self.current_lines = [line]
                else:
                    self.current_lines.append(line)

    def handles_last_call(self):
        return True


class RegexProcessor(ConfigProcessor):
    """Replaces all occurences of regex matches.

    Args:
        regexes (dict): a map of regexes and their replacements
    """

    def __init__(self, **init_params):

        super(RegexProcessor, self).__init__(**init_params)
        self.regexes = init_params["regexes"]

    def get_input_format(self):
        return STRING_FORMAT

    def process_current_config(self):
        new_config = self.current_input_config

        for regex, replacement in self.regexes.items():
            new_config = re.sub(regex, replacement, new_config)

        return new_config


class LoadMoreConfigsProcessor(ConfigProcessor):
    """Processort to load additional configs from configs.

    If an incoming configuration is a list of strings, it'll interprete it as list of
    urls and adds it in front of the list of 'yet-to-process' configs of this processing run.

    Use this with caution, since if this gets a list of string that is not a list of urls, it
    will still treat it like one and your run will fail.
    """

    def get_input_format(self):

        return PYTHON_FORMAT

    def get_output_format(self):

        return NONE_FORMAT

    def process_current_config(self):

        if is_list_of_strings(self.current_input_config):
            return None
        else:
            return self.current_input_config

    def get_additional_configs(self):

        if is_list_of_strings(self.current_input_config):
            return self.current_input_config
        else:
            return None


class UrlAbbrevProcessor(ConfigProcessor):
    """Replaces strings in an input configuration url with its expanded version.

    The default constructor without any arguments will create a processor only using the default, inbuilt abbreviations

    """

    def __init__(self, **init_params):

        super(UrlAbbrevProcessor, self).__init__(**init_params)

        abbrevs = init_params.get("abbrevs", False)
        add_default_abbrevs = init_params.get("add_default_abbrevs", True)

        if not abbrevs:
            if add_default_abbrevs:
                self.abbrevs = DEFAULT_URL_ABBREVIATIONS_FILE
            else:
                self.abbrevs = {}
        else:
            if add_default_abbrevs:
                self.abbrevs = copy.deepcopy(DEFAULT_URL_ABBREVIATIONS_FILE)
                dict_merge(self.abbrevs, abbrevs, copy_dct=False)
            else:
                self.abbrevs = copy.deepcopy(abbrevs)

        self.verbose = init_params.get("verbose", False)
        self.opt_split_string = "::"

    def get_input_format(self):

        return STRING_FORMAT

    def process_current_config(self):

        result = self.expand_config(self.current_input_config)
        return result

    def expand_config(self, config):
        """Expands abbreviated configuration urls that start with `<token>:`.

        This is a convenience for the user, as they don't have to type out long urls if they don't want to.

        To make it easier to remember (and shorter to type) config urls, *frkl*
        supports abbreviations which will be replaced before attempting to load a
        configuration. In addition to inbuild abbreviations, a custom ones can be
        piped into the Frkl constructor. This needs to be a dictionary of string to
        string (abbreviation -> full url-part, eg. ``freckles_configs -->
        https://raw.githubusercontent.com/makkus/freckles/master/examples``, which
        would enable the user to provide this config url:
        ``freckles_config:quickstart.yml``) or string to list (abbreviation -> list
        of tokens to assemble the finished string, which for example for getting
        raw files from the master branch of a github repo would look like:
        ["https://raw.githubusercontent.com", frutils.defaults.URL_PLACEHOLDER, frutils.defaults.URL_PLACEHOLDER,
        "master] -- the input config ``gh:makkus/freckles/examples/quickstart.yml``
        would result in the same string as in the first example above -- this
        method is a bit more fragile, because the input string can't have less
        tokens seperated by '/' than the value list).

        Args:
          config (str): the configuration url/json/etc...

        Returns:
          str: the configuration with all occurances of registered abbreviations replaced

        """

        # check if string contains branch information
        branch = None
        opt_split_string = "::"
        if opt_split_string in config:
            tokens = config.split(opt_split_string)
            opt = tokens[1:-1]
            if not opt:
                raise Exception(
                    "Not a valid url, needs at least 2 split strings ('{}')".format(
                        opt_split_string
                    )
                )
            if len(opt) != 1:
                raise Exception(
                    "Not a valid url, can only have 1 branch: {}".format(opt)
                )
            branch = opt[0]

        prefix, sep, rest = config.partition(":")

        if self.opt_split_string in rest:
            tokens = rest.split(self.opt_split_string)
            rest = "{}{}".format(tokens[0], tokens[-1])
            opt = tokens[1:-1]
            if not opt:
                raise Exception(
                    "Invalid url, need at least two option splitters ('{}'): {}".format(
                        self.opt_split_string, config
                    )
                )
        else:
            opt = None

        if prefix in self.abbrevs.keys():

            if isinstance(self.abbrevs[prefix], string_types):
                return "{}{}".format(self.abbrevs[prefix], rest)
            else:
                tokens = rest.split("/")
                tokens_copy = copy.copy(tokens)

                min_tokens = self.abbrevs[prefix].count(URL_PLACEHOLDER)

                result_string = ""
                for t in self.abbrevs[prefix]:

                    if t == URL_PLACEHOLDER:
                        if not tokens:
                            raise FrklConfigException(
                                "Can't expand url '{}': not enough parts, need at least {} parts seperated by '/' after ':'".format(
                                    config, min_tokens
                                )
                            )
                        to_append = tokens.pop(0)
                        if not to_append:
                            raise FrklConfigException(
                                "Last token empty, can't expand: {}".format(tokens_copy)
                            )
                    else:
                        to_append = t

                    result_string += to_append

                if tokens:
                    postfix = "/".join(tokens)
                    result_string += postfix

                # if opt:
                #    opt_appendix = "::".join(opt)

                if string_is_templated(result_string, DEFAULT_FRKL_JINJA_ENVIRONMENT):

                    if branch:
                        repl_dict = {"branch": branch}
                    else:
                        repl_dict = {"branch": "master"}

                    result_string = replace_string(
                        result_string,
                        repl_dict,
                        jinja_env=DEFAULT_FRKL_JINJA_ENVIRONMENT,
                        local_env_vars_key=False,
                    )

                if self.verbose:
                    print("Expanding '{}' -> '{}'".format(config, result_string))

                return result_string
        else:
            return config



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import io
import os

from frutils import calculate_cache_location_for_url, ensure_parent_dir
from frutils.defaults import DEFAULT_DOWNLOAD_CACHE_BASE
from ruamel.yaml import YAML
from six import string_types

from .frkl import Frkl
from .processors import EnsureUrlProcessor, UrlAbbrevProcessor


def get_full_url(abbrev_or_url, abbrevs=None):
    """Expands input to a full url if abbreviated.

    Args:
        abbrev_or_url (str): a full or abbreviated url

    Returns
        str: the full url
    """

    u = UrlAbbrevProcessor(abbrevs=abbrevs)
    chain = [u]

    single_input = False
    if isinstance(abbrev_or_url, string_types):
        single_input = True
        urls = [abbrev_or_url]

    f = Frkl(urls, chain)
    result = f.process()

    if single_input:
        if not result:
            return []
        return result[0]
    else:
        return result


def download_cached_file(
    abbrev_or_url,
    update=False,
    cache_base=DEFAULT_DOWNLOAD_CACHE_BASE,
    abbrevs=None,
    return_content=False,
):
    """Downloads a file using a full or abbreviated url.

    Args:
        abbrev_or_url (str): the full or abbreviated url
        update (bool): whether to 'force' update the file if it doesn't exist
        cache_base (str): root of cache directory
        abbrevs (dict): (optional) url abbreviations
        return_content (bool): whether to return the path to the file (False) or it's content (True)

    Returns:
        str: the path to the downloaded file or it's content (depending on the 'return_content' variable)
    """

    if not isinstance(abbrev_or_url, string_types):
        raise Exception("Url needs to be string")

    rel_cache_path = calculate_cache_location_for_url(get_full_url(abbrev_or_url))

    target_file = os.path.realpath(os.path.join(cache_base, rel_cache_path))
    if os.path.exists(target_file) and os.path.isfile(target_file) and not update:
        if return_content:
            with io.open(target_file, encoding="utf-8") as f:
                content = f.read()
                return content
        else:
            return target_file
    elif os.path.exists(target_file) and not os.path.isfile(target_file):
        raise Exception("Target file is not a file: {}".format(target_file))

    chain = [UrlAbbrevProcessor(abbrevs=abbrevs), EnsureUrlProcessor()]

    f = Frkl([abbrev_or_url], chain)
    result = f.process()[0]

    ensure_parent_dir(target_file)
    with io.open(target_file, "w", encoding="utf-8") as f:
        f.write(result)

    if return_content:
        return result
    else:
        return target_file


def content_from_url(
    abbrev_or_url, update=False, cache_base=DEFAULT_DOWNLOAD_CACHE_BASE, abbrevs=None
):
    """Downloads a file using a full or abbreviated url and returns it's content.

    Args:
        abbrev_or_url (str): the full or abbreviated url
        update (bool): whether to 'force' update the file if it doesn't exist
        cache_base (str): root of cache directory
        abbrevs (dict): (optional) url abbreviations

    Returns:
        str: the content of the downloaded file
    """

    content = download_cached_file(
        abbrev_or_url,
        update=update,
        cache_base=cache_base,
        abbrevs=abbrevs,
        return_content=True,
    )
    return content


def dict_from_url(
    abbrev_or_url, update=False, cache_base=DEFAULT_DOWNLOAD_CACHE_BASE, abbrevs=None
):
    """Downloads a file using a full or abbreviated url and returns it's content as a python dict.

    Args:
        abbrev_or_url (str): the full or abbreviated url
        update (bool): whether to 'force' update the file if it doesn't exist
        cache_base (str): root of cache directory
        abbrevs (dict): (optional) url abbreviations

    Returns:
        str: the content of the downloaded file
    """

    content = download_cached_file(
        abbrev_or_url,
        update=update,
        cache_base=cache_base,
        abbrevs=abbrevs,
        return_content=True,
    )
    yaml = YAML(typ="safe")

    result = yaml.load(content)
    return result



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

from frutils.defaults import JINJA_DELIMITER_PROFILES

from .defaults import *  # noqa: F403
from .processors import (
    EnsurePythonObjectProcessor,
    EnsureUrlProcessor,
    FrklProcessor,
    Jinja2TemplateProcessor,
    UrlAbbrevProcessor,
)

# simple chain to convert a string (which might be an abbreviated url or path or yaml or json string) into a python object
LOAD_OBJECT_FROM_URL_CHAIN = [
    UrlAbbrevProcessor(),
    EnsureUrlProcessor(),
    EnsurePythonObjectProcessor(),
]

LOAD_STRING_FROM_URL_CHAIN = [UrlAbbrevProcessor(), EnsureUrlProcessor()]


# format of processor init dicts
BOOTSTRAP_FRKL_FORMAT = {
    STEM_KEY_NAME: "processors",
    DEFAULT_LEAF_KEY_NAME: "processor",
    DEFAULT_LEAF_DEFAULT_KEY_NAME: "type",
    OTHER_VALID_KEYS_NAME: ["init"],
    DEFAULT_LEAF_KEY_MAP_NAME: "init",
}

COLLECTOR_INIT_BOOTSTRAP_PROCESSOR_CHAIN = [FrklProcessor(**BOOTSTRAP_FRKL_FORMAT)]

# chain to bootstrap processor_chain in order to generate a frkl object
BOOTSTRAP_PROCESSOR_CHAIN = [
    UrlAbbrevProcessor(),
    EnsureUrlProcessor(),
    EnsurePythonObjectProcessor(),
    FrklProcessor(**BOOTSTRAP_FRKL_FORMAT),
]


def load_templated_string_from_url_chain(
    repl_dict,
    create_python_object=False,
    use_environment_vars=False,
    use_context=False,
    delimiter_profile=JINJA_DELIMITER_PROFILES["default"],
):
    """Assemles a chain to load a file (local or remote) and replace templated values in it's content.

     Args:
        template_values (dict): a dictionary containing the values to replace template strings with
        create_python_object (bool): whether to return the string (False), or try to convert into python object (True)
        use_environment_vars (bool): whether to also use current environment variables  in the replacement dict. If True, it'll stored under the key 'LOCAL_ENV', if string, that will be used as key
        use_context (bool): whether to also use the frkl context in the replacement dict.  If True, it'll stored under the key 'frkl_vars', if string, that will be used as key
        delimiter_profile (dict): the jinja2 delimters

    Returns:
        str: the result string
    """

    template_processor = Jinja2TemplateProcessor(
        repl_dict,
        use_environment_vars=use_environment_vars,
        use_context=use_context,
        delimiter_profile=delimiter_profile,
    )
    chain = LOAD_STRING_FROM_URL_CHAIN + [template_processor]
    if create_python_object:
        chain = chain + [EnsurePythonObjectProcessor()]

    return chain



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

from .callbacks import *  # noqa: F403
from .chains import *  # noqa: F403
from .frkl import Frkl  # noqa: F403
from .processors import *  # noqa: F403

__metaclass__ = type

log = logging.getLogger("frkl")


def init(files_or_folders, additional_configs=None, use_strings_as_config=False):
    """Creates a Frkl object.

        Args:
          files_or_folders (list): a list of files or folders or url strings. if the item is a file or url string, it will be used as Frkl bootstrap config, if it is a folder, it is forwarded to the 'from_folder' method to get lists of bootstrap and/or config files
          additional_configs (list): a list of files or url strings, used as configs for the initialized Frlk object
          use_strings_as_config (bool): whether to use non-folder strings as config files (instead of to initialze the Frkl object, which is default)

        Returns:
          Frkl: the object
        """

    if additional_configs is None:
        additional_configs = []
    chain_files = []
    config_files = []

    for f in files_or_folders:
        if not os.path.exists(f):
            # means this is a url string
            if not use_strings_as_config:
                chain_files.append(f)
            else:
                config_files.append(f)
        elif os.path.isfile(f):
            # means we can use this directly
            chain_files.append(f)
        else:
            temp_chain, temp_config = get_configs(f)
            chain_files.extend(temp_chain)
            config_files.extend(temp_config)

    if not chain_files:
        raise FrklConfigException(
            "No bootstrap information for Frkl found, can't create object."
        )

    frkl_obj = factory(chain_files, config_files)
    return frkl_obj


def from_folder(folders):
    """Creates a Frkl object using a folder path as the only input.

        The folder needs to contain one or more files that start with the '_' and end with '.yml',
        which are used to bootstrap the frkl object by reading them in alphabetical order,
        and one or more additional files with the 'yml' extension, which are then used as
        input configurations, again in alphabetical order.

        Args:
          folders (list): paths to local folder(s)

        Returns:
          Frkl: the initialized Frkl object
        """

    chain_files, config_files = get_configs(folders)

    if not chain_files:
        raise FrklConfigException(
            "No bootstrap information for Frkl found, can't create object."
        )

    frkl_obj = factory(chain_files, config_files)
    return frkl_obj


def get_configs(folders):
    """Looks at a folder and retrieves configs.

        The folders need to contain one or more files that start with the '_' and end with '.yml',
        which are used to bootstrap the frkl object by reading them in alphabetical order,
        and one or more additional files with the 'yml' extension, which are then used as
        input configurations, again in alphabetical order.

        Args:
          folders (list): paths to one or several local folders

        Returns:
          tuple: first element of the tuple is a list of bootstrap configurations, 2nd element is a list of actual configs
        """

    if isinstance(folders, string_types):
        folders = [folders]

    all_chains = []
    all_configs = []
    for folder in folders:
        chain_files = []
        config_files = []
        for child in os.listdir(folder):
            if (
                not child.startswith("__")
                and child.startswith("_")
                and child.endswith(".yml")
            ):
                chain_files.append(os.path.join(folder, child))
            elif child.endswith(".yml"):
                config_files.append(os.path.join(folder, child))

        chain_files.sort()
        config_files.sort()

        all_chains.extend(chain_files)
        all_configs.extend(config_files)

    return (all_chains, all_configs)


def factory(bootstrap_configs, frkl_configs=None):
    """Factory method to easily create a Frkl object using a list of configurations to describe
        the format of the configs to use later on, as well as (optionally) a list of such configs.

        Args:
          bootstrap_configs (list): the configuration to describe the format of the configurations the new Frkl object uses
          frkl_configs (list): (optional) configurations to init the Frkl object with. this can also be done later using the 'set_configs' method

        Returns:
          Frkl: a new Frkl object
        """

    if frkl_configs is None:
        frkl_configs = []
    if isinstance(bootstrap_configs, string_types):
        bootstrap_configs = [bootstrap_configs]

    bootstrap = Frkl(bootstrap_configs, BOOTSTRAP_PROCESSOR_CHAIN)
    config_frkl = bootstrap.process(FrklFactoryCallback())

    config_frkl.set_configs(frkl_configs)
    return config_frkl



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for frkl.
    Use setup.cfg to configure your project.
"""

import sys
from pkg_resources import VersionConflict, require
from setuptools import setup

try:
    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


if __name__ in ["__main__", "builtins", "__builtin__"]:
    setup(use_scm_version={"write_to": "src/frkl/version.txt"})



# -*- coding: utf-8 -*-

# python 3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

# ------------------------------------------------------------------------
# Frkl Exception(s)


class FrklConfigException(Exception):
    def __init__(self, message, errors=None):
        """Exception that is thrown when processing configuration urls/content.

        Args:
          message (str): the error message
          errors (list): list of root causes and error descriptions
        """

        super(FrklConfigException, self).__init__(message)
        if errors is None:
            errors = []
        if isinstance(errors, Exception):
            self.errors = [errors]
        else:
            self.errors = errors


class FrklistException(Exception):
    """Base exception class for nsbl."""

    def __init__(self, message):

        super(FrklistException, self).__init__(message)


class FrklistConfigException(FrklistException):
    def __init__(self, message):

        super(FrklConfigException, self).__init__(message)
